"""
Dialog functions used throughout the GUI.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtWidgets import (
    QVBoxLayout,
    QDialog,
    QLabel,
    QDialogButtonBox,
    QComboBox,
    QFormLayout,
)

from ..widgets import get_widget_value, create_setting_widget


class OperationDialog(QDialog):
    def __init__(self, operation_type, parameters, parent=None):
        super().__init__(parent)
        self.operation_type = operation_type
        self.parameters = parameters
        self.parameter_widgets = {}
        self.label_widgets = {}
        self.is_hierarchical = isinstance(parameters, dict)

        self.setWindowTitle(self.operation_type)
        self.main_layout = QVBoxLayout(self)
        self.params_layout = QFormLayout()

        if self.is_hierarchical:
            self.type_selector = QComboBox()
            self.type_selector.addItems(list(self.parameters.keys()))
            self.type_selector.currentIndexChanged.connect(
                self.update_operation_options
            )
            self.params_layout.addRow("Option:", self.type_selector)

        self.main_layout.addLayout(self.params_layout)

        button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        self.main_layout.addWidget(button_box)

        if self.is_hierarchical:
            return self.update_operation_options(0)
        return self.update_parameters(self.parameters)

    def update_operation_options(self, index):
        current_type = list(self.parameters.keys())[index]
        self.update_parameters(self.parameters[current_type])

    def update_parameters(self, parameters):
        while self.params_layout.rowCount() > (1 if self.is_hierarchical else 0):
            self.params_layout.removeRow(self.params_layout.rowCount() - 1)
        self.parameter_widgets.clear()
        self.label_widgets.clear()

        for param_info in parameters:
            label, value, min_value, tooltip_info = param_info

            settings = {
                "label": tooltip_info.get("title"),
                "description": tooltip_info.get("description"),
                "notes": tooltip_info.get("notes"),
                "default": value,
                "min": min_value,
            }

            type = "number"
            if isinstance(value, bool):
                type = "boolean"
            elif isinstance(min_value, list):
                type = "select"
                settings["options"] = min_value
            elif isinstance(value, (float, str)):
                type = "text"

            settings["type"] = type
            widget = create_setting_widget(settings)
            label_widget = QLabel(f"{tooltip_info['title']}:")
            self.label_widgets[label] = label_widget
            self.parameter_widgets[label] = widget
            self.params_layout.addRow(label_widget, widget)

    def get_parameters(self):
        ret = {}
        for param_name, widget in self.parameter_widgets.items():
            ret[param_name] = get_widget_value(widget)
        return ret


class ParameterHandler:
    def __init__(self, operation_dict, settings_button, selector):
        self.operation_dict = operation_dict
        self.settings_button = settings_button
        self.selector = selector
        self.parameters_store = {}

        self.update_button(self.selector.currentText())
        for op_type, params in self.operation_dict.items():
            if not params:
                continue
            self.parameters_store[op_type] = {x[0]: x[1] for x in params}

    def get(self, key, default=None):
        return self.parameters_store.get(key, default)

    def update_button(self, current_type):
        has_params = len(self.operation_dict[current_type]) > 0
        self.settings_button.setEnabled(has_params)

    def show_dialog(self):
        current_type = self.selector.currentText()
        params = []
        custom_parameters = self.parameters_store.get(current_type, {})

        for param in self.operation_dict[current_type].copy():
            default = custom_parameters.get(param[0], param[1])
            params.append(tuple(x if i != 1 else default for i, x in enumerate(param)))

        def _ident(**kwargs):
            return kwargs

        kwargs = show_parameter_dialog(
            current_type,
            params,
            self.settings_button.parent(),
            {current_type: _ident},
            self.settings_button,
        )

        if isinstance(kwargs, dict):
            self.parameters_store[current_type] = kwargs


def show_parameter_dialog(
    operation_type, parameters, obj, mapping={}, source_widget=None
):
    dialog = OperationDialog(operation_type, parameters, obj)
    if source_widget:
        pos = source_widget.mapToGlobal(source_widget.rect().bottomLeft())
        dialog.move(pos)

    if dialog.exec() == QDialog.DialogCode.Rejected:
        return -1

    params = dialog.get_parameters()
    if dialog.is_hierarchical:
        params["method"] = dialog.type_selector.currentText()

    func = mapping.get(operation_type)
    if func is None:
        print(f"{operation_type} is unknown - Supported are {mapping.keys()}.")
    return func(**params)


def make_param(param, default, min_val=0, description=None, notes=None):
    return (
        param,
        default,
        min_val,
        {
            "title": param.lower().replace("_", " ").capitalize(),
            "description": description or param,
            "default_value": str(default),
            "notes": notes if notes else None,
        },
    )



from typing import Tuple, Dict
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QFrame,
    QScrollArea,
    QWidget,
    QGroupBox,
    QGridLayout,
)
import qtawesome as qta

from ..widgets import DialogFooter
from ..stylesheets import QGroupBox_style, QPushButton_style, QScrollArea_style
from ..widgets import create_setting_widget, get_widget_value


class StyleableButton(QPushButton):
    def __init__(
        self, icon_name, title, description=None, is_compact=False, parent=None
    ):
        super().__init__(parent)

        layout = QVBoxLayout(self)

        icon_size = 32
        size = (150, 100)
        margin = 8, 12, 8, 12
        if is_compact:
            icon_size = 24
            size = (70, 70)
            margin = (6, 8, 6, 8)

        layout.setContentsMargins(*margin)
        layout.setSpacing(4)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)

        icon = qta.icon(icon_name, color="#696c6f")
        icon_label = QLabel()
        icon_label.setPixmap(icon.pixmap(icon_size, icon_size))
        icon_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(icon_label)

        title_label = QLabel(title)
        title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(title_label)

        if description and not is_compact:
            desc_label = QLabel(description)
            desc_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
            desc_label.setStyleSheet("color: #696c6f; font-size: 11px;")
            desc_label.setWordWrap(True)
            layout.addWidget(desc_label)

        self.setMinimumSize(*size)
        self.setCheckable(True)
        self.setStyleSheet(
            """
            QPushButton {
                border: 1px solid #d1d5db;
                border-radius: 6px;
                text-align: center;
            }
            QPushButton:checked {
                border: 1px solid #4f46e5;
            }
            QPushButton:hover:!checked {
                background-color: #1a000000;
            }
        """
        )


class ExportDialog(QDialog):
    export_requested = Signal(dict)

    def __init__(self, parent=None, parameters={}, enabled_categories=None):
        super().__init__(parent)

        self.setWindowTitle("Export Data")
        self.resize(700, 600)

        if enabled_categories is None:
            enabled_categories = ["pointcloud", "mesh", "volume"]
        self.enabled_categories = set(enabled_categories)
        self.format_categories = {
            "pointcloud": {
                "icon": "mdi.dots-grid",
                "label": "Point Cloud",
                "description": "Export coordinates and orientations.",
                "formats": ["star", "tsv", "xyz"],
            },
            "mesh": {
                "icon": "mdi.grid",
                "label": "Mesh",
                "description": "Export as a surface mesh.",
                "formats": ["obj", "stl", "ply"],
            },
            "volume": {
                "icon": "mdi.cube-outline",
                "label": "Volume",
                "description": "Export as a density map.",
                "formats": ["mrc", "em", "h5"],
            },
        }

        self.format_settings_definitions = {
            "mrc": volume_settings,
            "em": volume_settings,
            "h5": volume_settings,
            "xyz": {
                "single_file": {
                    "type": "boolean",
                    "label": "Single File",
                    "description": "Export all data to a single file",
                    "default": False,
                    "parameter": "single_file",
                },
                "header": {
                    "type": "boolean",
                    "label": "Include Header",
                    "description": "Include column headers in the exported file",
                    "default": True,
                    "parameter": "header",
                },
            },
            "star": {
                "single_file": {
                    "type": "boolean",
                    "label": "Single File",
                    "description": "Export all data to a single file",
                    "default": False,
                    "parameter": "single_file",
                },
                "relion_5_format": {
                    "type": "boolean",
                    "label": "RELION 5",
                    "description": "Export in RELION 5 format with coordinate transformation",
                    "default": False,
                    "parameter": "relion_5_format",
                },
                "shape_x": {
                    "type": "number",
                    "label": "Shape X",
                    "description": "X voxel for coordinate transformation (RELION 5)",
                    "default": 64,
                    "min": 1,
                    "parameter": "shape_x",
                },
                "shape_y": {
                    "type": "number",
                    "label": "Shape Y",
                    "description": "Y voxel for coordinate transformation (RELION 5)",
                    "default": 64,
                    "min": 1,
                    "parameter": "shape_y",
                },
                "shape_z": {
                    "type": "number",
                    "label": "Shape Z",
                    "description": "Z voxel for coordinate transformation (RELION 5)",
                    "default": 64,
                    "min": 1,
                    "parameter": "shape_z",
                },
            },
            "tsv": {
                "single_file": {
                    "type": "boolean",
                    "label": "Single File",
                    "description": "Export all data to a single file",
                    "default": False,
                    "parameter": "single_file",
                }
            },
            "obj": {},
            "stl": {},
            "ply": {},
        }

        self.selected_category = next(
            (
                cat
                for cat in ["pointcloud", "mesh", "volume"]
                if cat in self.enabled_categories
            ),
            "pointcloud",
        )
        self.selected_format = self.format_categories[self.selected_category][
            "formats"
        ][0]

        self.selected_format = "star"
        self.current_settings = {}
        self.show_advanced = False

        # Set parameters before drawing dialog
        self.set_defaults(list(parameters.keys()), list(parameters.values()))

        self.setup_ui()
        self.setStyleSheet(QGroupBox_style + QPushButton_style + QScrollArea_style)

    def set_defaults(self, keys, values):
        """Update default values for format settings"""
        for format_name, settings_dict in self.format_settings_definitions.items():
            for index, key in enumerate(keys):
                if key in settings_dict:
                    settings_dict[key]["default"] = values[index]

    def setup_ui(self):
        from ..icons import icon_color

        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

        # Scroll area for content
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFrameShape(QFrame.Shape.NoFrame)

        content_widget = QWidget()
        scroll_area.setWidget(content_widget)
        content_layout = QVBoxLayout(content_widget)
        content_layout.setContentsMargins(20, 20, 20, 20)
        content_layout.setSpacing(20)

        export_group = QGroupBox("Export Type")
        self.export_layout = QHBoxLayout(export_group)
        self.setup_group_buttons()
        content_layout.addWidget(export_group)

        format_group = QGroupBox("File Format")
        self.format_layout = QHBoxLayout(format_group)
        self.setup_format_buttons()
        content_layout.addWidget(format_group)

        settings_group = QGroupBox("Settings")
        settings_group.setMinimumHeight(200)
        self.settings_layout = QVBoxLayout(settings_group)
        self.update_advanced_settings()
        content_layout.addWidget(settings_group)

        main_layout.addWidget(scroll_area)

        footer = DialogFooter(dialog=self, margin=(20, 10, 20, 10))
        footer.accept_button.setText("Export")
        footer.accept_button.setIcon(qta.icon("fa5s.download", color=icon_color))
        main_layout.addWidget(footer)

    def _clear_layout(self, layout):
        while layout.count():
            item = layout.takeAt(0)
            if item.widget():
                item.widget().deleteLater()
        return layout

    def setup_group_buttons(self):
        self._clear_layout(self.export_layout)
        self.category_buttons = {}
        for i, (cat_id, category) in enumerate(self.format_categories.items()):
            btn = StyleableButton(
                category["icon"],
                category["label"],
                category["description"],
                is_compact=False,
            )

            is_enabled = cat_id in self.enabled_categories
            btn.setEnabled(is_enabled)
            btn.setChecked(cat_id == self.selected_category and is_enabled)

            if is_enabled:
                btn.clicked.connect(
                    lambda checked, cat=cat_id: self.on_category_selected(cat)
                )

            self.export_layout.addWidget(btn)
            self.category_buttons[cat_id] = btn

    def setup_format_buttons(self):
        self._clear_layout(self.format_layout)
        self.format_buttons = {}
        formats = self.format_categories[self.selected_category]["formats"]

        for i, fmt in enumerate(formats):
            btn = StyleableButton("fa5s.file", f".{fmt}", is_compact=True)
            btn.setChecked(fmt == self.selected_format)
            btn.clicked.connect(lambda checked, f=fmt: self.on_format_selected(f))
            self.format_layout.addWidget(btn)
            self.format_buttons[fmt] = btn

    def update_advanced_settings(self):
        self._clear_layout(self.settings_layout)

        settings_definitions = self.format_settings_definitions.get(
            self.selected_format, {}
        )

        if not settings_definitions:
            self.settings_grid_layout = None
            no_settings_label = QLabel(
                "No additional settings available for this format."
            )
            no_settings_label.setStyleSheet("color: #6b7280; font-style: italic;")
            self.settings_layout.addWidget(no_settings_label)
            return

        settings_widget = QWidget()
        grid_layout = QGridLayout(settings_widget)
        grid_layout.setContentsMargins(10, 10, 10, 10)
        grid_layout.setSpacing(10)

        row = 0
        col = 0
        for setting_key, setting_def in settings_definitions.items():
            widget = create_setting_widget(setting_def)

            label = QLabel(setting_def["label"])
            grid_layout.addWidget(label, row, col * 2)
            grid_layout.addWidget(widget, row, col * 2 + 1)

            col = 1 - col
            if col == 0:
                row += 1

        self.settings_layout.addWidget(settings_widget)
        self.settings_grid_layout = grid_layout

    def get_current_settings(self) -> Dict:
        """Extract current settings from the grid widgets"""
        settings = {}

        if getattr(self, "settings_grid_layout", None) is None:
            return settings

        # Collect values from all widgets in the grid
        for i in range(self.settings_grid_layout.count()):
            item = self.settings_grid_layout.itemAt(i)
            if item and item.widget():
                widget = item.widget()
                parameter = widget.property("parameter")
                if not parameter:
                    continue
                settings[parameter] = get_widget_value(widget)

        return settings

    def on_category_selected(self, category_id):
        if category_id == self.selected_category:
            return None

        if category_id not in self.enabled_categories:
            return None

        for cat_id, btn in self.category_buttons.items():
            btn.setChecked(cat_id == category_id)

        self.selected_category = category_id
        self.selected_format = self.format_categories[category_id]["formats"][0]
        self.current_settings = {}

        self.setup_format_buttons()
        self.update_advanced_settings()

    def on_format_selected(self, format_id):
        if format_id == self.selected_format:
            return None

        for fmt, btn in self.format_buttons.items():
            btn.setChecked(fmt == format_id)

        self.selected_format = format_id
        self.current_settings = {}
        self.update_advanced_settings()

    def accept(self):
        export_data = {
            "category": self.selected_category,
            "format": self.selected_format,
            **self.get_current_settings(),
        }

        self.export_requested.emit(export_data)
        return super().accept()


volume_settings = {
    "shape_x": {
        "type": "number",
        "label": "Shape X",
        "description": "X dimension of the volume",
        "default": 64,
        "min": 1,
        "parameter": "shape_x",
    },
    "shape_y": {
        "type": "number",
        "label": "Shape Y",
        "description": "Y dimension of the volume",
        "default": 64,
        "min": 1,
        "parameter": "shape_y",
    },
    "shape_z": {
        "type": "number",
        "label": "Shape Z",
        "description": "Z dimension of the volume",
        "default": 64,
        "min": 1,
        "parameter": "shape_z",
    },
    "sampling": {
        "type": "float",
        "label": "Sampling Rate",
        "description": "Sampling rate in Ångströms",
        "notes": "Defaults to sampling rate of Geometry object",
        "default": -1,
        "min": -1,
        "step": 0.1,
        "decimals": 8,
        "parameter": "sampling",
    },
}



import textwrap
import warnings
from typing import Dict, List
from os.path import join, basename

import numpy as np

from ..utils import find_closest_points

from . import remesh, center_mesh, compute_scale_factor_lower, scale
from ..formats.writer import write_topology_file

__all__ = ["mesh_to_cg"]


def mesh_to_cg(
    mesh,
    output_directory: str,
    inclusions: List[Dict],
    edge_length: float = 40.0,
    include_normals: bool = True,
    flip_normals: bool = True,
) -> bool:
    from ..parametrization import TriangularMesh

    mesh = remesh(mesh, edge_length)

    vertex_maps = []
    for protein in inclusions:
        geometry = protein.get("data", None)
        if geometry is None:
            continue

        if include_normals:
            fit = TriangularMesh(mesh)
            kwargs = {
                "points": geometry.points,
                "normals": geometry.normals * -1 if flip_normals else 1,
                "return_indices": True,
            }
            _, vertex_indices = fit.compute_distance(**kwargs)
        else:
            _, vertex_indices = find_closest_points(mesh.vertices, geometry.points, k=1)

        vertex_indices, incl_indices = np.unique(vertex_indices, return_index=True)

        incl_quat = geometry.quaternions[incl_indices]
        vertex_maps.append([vertex_indices, incl_indices, incl_quat])

    n_inclusions = len(vertex_maps)
    for i in range(n_inclusions):
        for k in range(i + 1, n_inclusions):
            overlap = np.intersect1d(vertex_maps[i][0], vertex_maps[k][0])
            if len(overlap) == 0:
                continue

            keep = np.invert(np.isin(vertex_maps[k][0], overlap))
            for ix in range(len(vertex_maps[k])):
                vertex_maps[k][ix] = vertex_maps[k][ix][keep]

            print(
                f"{inclusions[i]['name']} and {inclusions[k]['name']} collide on "
                f"{len(overlap)} vertices. Consider reducing mesh edge length. "
                f"Continuing but dropping collisions from {inclusions[k]['name']}."
            )

    scale_factor = compute_scale_factor_lower(mesh, lower_bound=1.0)
    mesh_scale = scale(mesh, scale_factor)
    data, offset = center_mesh(mesh_scale)
    offset = ",".join([str(-float(x)) for x in offset])

    inclusion_list, inclusion_quat = [], []
    for index, ret in enumerate(vertex_maps):
        inclusion_list.extend([(index + 1, x, 0, 1) for x in ret[0]])
        inclusion_quat.extend([(index + 1, x, y, *z) for x, y, z in zip(*ret)])

    _inclusions = np.zeros((len(inclusion_list), 5))
    _inclusions[:, 0] = np.arange(_inclusions.shape[0])
    _inclusions[:, 1:5] = np.asarray(inclusion_list)
    data["inclusions"] = _inclusions.astype(int)

    mesh_path = join(output_directory, "mesh.tsi")
    write_topology_file(file_path=mesh_path, data=data, tsi_format=True)

    normal_path = join(output_directory, "inclusion_quaternions.csv")
    inclusion_quat = np.asarray(inclusion_quat)
    np.savetxt(
        normal_path,
        inclusion_quat,
        delimiter=",",
        comments="",
        header="inclusion_type,vertex,inclusion,w,x,y,z",
    )

    warnings.warn(
        "Setup TS2CG - Corresponding Citation: "
        "[1] Pezeshkian, W. et al. (2020) Nat. Commun., doi.org/10.1038/s41467-020-16094-y."
    )
    scale_path = join(output_directory, "scales.txt")
    with open(scale_path, mode="w", encoding="utf-8") as ofile:
        ofile.write(f"{mesh_path}\t{scale_factor}\t{offset}\n")

    str_path = join(output_directory, "input.str")
    with open(str_path, mode="w", encoding="utf-8") as ofile:
        for inclusion in inclusions:
            ofile.write(f"include {inclusion['name']}.gro\n")

        ofile.write("[Lipids List]\n")
        ofile.write("Domain 0\n")
        ofile.write("POPC 1 1 0.64\n")
        ofile.write("End\n")

        ofile.write("[Protein List]\n")
        for index, inclusion in enumerate(inclusions):
            ofile.write(f"{inclusion['name']} {index + 1} 0.01 0 0 0.0\n")
        ofile.write("End Protein\n")

    # Mosaic assumes angstrom for all spatial units
    mesh_to_nm = 0.1 / scale_factor

    plm_path = join(output_directory, "plm.sh")
    with open(plm_path, mode="w", encoding="utf-8") as ofile:
        plm_script = textwrap.dedent(
            f"""
            #!/bin/bash

            TS2CG PLM \\
                -TSfile {basename(mesh_path)} \\
                -bilayerThickness 3.8 \\
                -rescalefactor {mesh_to_nm} {mesh_to_nm} {mesh_to_nm}
        """
        )
        ofile.write(plm_script.lstrip())

    pcg_path = join(output_directory, "pcg.sh")
    with open(pcg_path, mode="w", encoding="utf-8") as ofile:
        pcg_script = textwrap.dedent(
            f"""
            #!/bin/bash

            TS2CG PCG \\
                -str {basename(str_path)} \\
                -Bondlength 0.1 \\
                -LLIB martini3.LIB \\
                -defout system
        """
        )
        ofile.write(pcg_script.lstrip())

    martinize_path = join(output_directory, "martinize.sh")
    with open(martinize_path, mode="w", encoding="utf-8") as ofile:
        martinize_script = "#!/bin/bash\n"

        martinize_execute = ""
        for inclusion in inclusions:
            inclusion_var = f"{inclusion['name']}_structure".upper()

            martinize_script += f"{inclusion_var}=\n"
            martinize_execute += textwrap.dedent(
                f"""
                martinize2 \\
                    -f ${inclusion_var} \\
                    -x {inclusion['name']}.pdb \\
                    -o {inclusion['name']}.top \\
                    -p backbone \\
                    -elastic \\
                    -maxwarn 2 \\
                    -merge all
            """
            )
        martinize_script += martinize_execute
        ofile.write(martinize_script)

    return True
